(*

fsharpi Mini.fs Algorithmics/Sort.fs Algorithmics/Area.fs

open Algorithmics.Area ;;

*)

module Algorithmics.Area

open Mini

open Algorithmics.Sort

///////////////////////////////////////////////////////////////////////////////
// Intervalle.

type Interval  =  { lo : int; hi : int }

module I =

    let length (i : Interval)  =
        max 0 (i.hi - i.lo)

//    let intersection (i : Interval, j : Interval)  =
//        { lo = max i.lo j.lo; hi = min i.hi j.hi }

    let union (i : Interval, j : Interval)  =
        { lo = min i.lo j.lo; hi = max i.hi j.hi }

module Is =

  let rec union = function
      | ([], is) | (is, [])                                   ->  is
      | (i :: is, j :: js) when i.hi < j.lo                   ->  i :: union (is, j :: js)
      | (i :: is, j :: js) when j.hi < i.lo                   ->  j :: union (i :: is, js)
      | (i :: is, j :: js) when i.lo <  j.lo && i.hi <  j.hi  ->  union (is, I.union (i, j) :: js)
      | (i :: is, j :: js) when i.lo <= j.lo && i.hi >= j.hi  ->  union (i :: is, js)
      | (i :: is, j :: js) when i.lo >= j.lo && i.hi <= j.hi  ->  union (is, j :: js)
      | (i :: is, j :: js) when i.lo >  j.lo && i.hi >  j.hi  ->  union (I.union (i, j) :: is, js)

  // val intervals : Interval list -> Interval list
  let rec intervals = function
      | []   ->  []
      | [i]  ->  [i]
      | is   ->  let (is1, is2) = unzip is
                 union (intervals is1, intervals is2)

  let totalLength  =
      List.sumBy I.length

///////////////////////////////////////////////////////////////////////////////
// Rechtecke.

type Rectangle  =  { x : Interval; y : Interval }

module R =

    let area (r : Rectangle)  =
        I.length r.x * I.length r.y

//    let intersection (r : Rectangle, s : Rectangle)  =
//        { x = I.intersection (r.x, s.x);
//          y = I.intersection (r.y, s.y) }

///////////////////////////////////////////////////////////////////////////////
// Streifen.

let splitAt x i  =
    if i.lo = x then error "aaarg lo"
    if i.hi = x then error "aaarg hi"
    ({ lo = i.lo; hi = x }, { lo = x; hi = i.hi })

let rec unionBy f = function
    | ([], is) | (is, [])
      ->  is
    | ((i, a) :: is, (j, b) :: js) when i.hi <=  j.lo
      ->  (i, a) :: unionBy f (is, (j, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when j.hi <=  i.lo
      ->  (j, b) :: unionBy f ((i, a) :: is, js)
    | ((i, a) :: is, (j, b) :: js) when i.lo <  j.lo
      ->  let (il, ir) = splitAt j.lo i in (il, a) :: unionBy f ((ir, a) :: is, (j, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when i.lo >  j.lo
      ->  let (jl, jr) = splitAt i.lo j in (jl, b) :: unionBy f ((i, a) :: is, (jr, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when i.lo =  j.lo && i.hi <  j.hi
      ->  let (jl, jr) = splitAt i.hi j in (i, f (a, b)) :: unionBy f (is, (jr, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when i.lo =  j.lo && i.hi =  j.hi
      ->  (i, f (a,b)) :: unionBy f (is, js)
    | ((i, a) :: is, (j, b) :: js) when i.lo =  j.lo && i.hi >  j.hi
      ->  let (il, ir) = splitAt j.hi i in (j, f (a, b)) :: unionBy f ((ir, a) :: is, js)

let rec rectangles0 = function
    | []   ->  []
    | [r]  ->  [(r.x, [r.y])]
//    | [r]  ->  [(r.y, [r.x])]
    | rs   ->  let (rs1, rs2) = unzip rs
               unionBy Is.union (rectangles0 rs1, rectangles0 rs2)

let rec Union = function
    | ([], is) | (is, [])
      ->  is
    | ((i, a) :: is, (j, b) :: js) when i.hi <=  j.lo
      ->  (i, a) :: Union (is, (j, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when j.hi <=  i.lo
      ->  (j, b) :: Union ((i, a) :: is, js)
    | ((i, a) :: is, (j, b) :: js) when i.lo <  j.lo
      ->  let (il, ir) = splitAt j.lo i in (il, a) :: Union ((ir, a) :: is, (j, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when i.lo >  j.lo
      ->  let (jl, jr) = splitAt i.lo j in (jl, b) :: Union ((i, a) :: is, (jr, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when i.lo =  j.lo && i.hi <  j.hi
      ->  let (jl, jr) = splitAt i.hi j in (i, a @ b) :: Union (is, (jr, b) :: js)
    | ((i, a) :: is, (j, b) :: js) when i.lo =  j.lo && i.hi =  j.hi
      ->  (i, a @ b) :: Union (is, js)
    | ((i, a) :: is, (j, b) :: js) when i.lo =  j.lo && i.hi >  j.hi
      ->  let (il, ir) = splitAt j.hi i in (j, a @ b) :: Union ((ir, a) :: is, js)

let rec unionAll = function
    | []   ->  []
    | [r]  ->  [(r.x, [r.y])]
//    | [r]  ->  [(r.y, [r.x])]
    | rs   ->  let (rs1, rs2) = unzip rs
               Union (unionAll rs1, unionAll rs2)

let rectangles rs =
    [ for (i, js) in unionAll rs -> (i, Is.intervals js) ] 

let totalArea  =
    List.sumBy (fun (i, js) -> I.length i * Is.totalLength js)

///////////////////////////////////////////////////////////////////////////////
// example data

let random seed n =
    let rnd = System.Random seed
    [ for i in 1 .. n ->
        let l = rnd.Next(0, 64)
        let b = rnd.Next(0, 35)
        { x = { lo = l; hi = l + rnd.Next(1, 15) };
          y = { lo = b; hi = b + rnd.Next(1, 8) }} ]

let campus  =
    [ for r in random 2765 50 ->
      { x = { lo = r.x.lo + 2; hi = r.x.hi + 2 };
        y = { lo = r.y.lo + 1; hi = r.y.hi + 1 }}]

let staircase n =
    [ for i in 1 .. n ->
      { x = { lo = 0; hi = i };
        y = { lo = i - 1; hi = i }} ]
        
let staircase2 n =
    [ for i in 1 .. n ->
      { x = { lo = 0; hi = i };
        y = { lo = 2 * i - 2; hi = 2 * i - 1 }} ]

///////////////////////////////////////////////////////////////////////////////
// Wie viele verschiedene Schnittmengen lassen sich mit n Rechtecken erzeugen?

let rects m n  =
    seq {
        for l in 0 .. m - 2 do
          for r in l + 2 .. m do
            for b in 0 .. n - 2 do
              for t in b + 2 .. n do
                yield { x = { lo = l; hi = r }; y = { lo = b; hi = t } }
    }

let contains (i, j) r  =  r.x.lo <= i && i <= r.x.hi && r.y.lo <= j && j <= r.y.hi

let regions m n rs  =
    let is =
        [ for i in 0 .. m do
            for j in 0 .. n do
              yield [ for r in rs -> contains (i,j) r ]
        ]
    List.length (List.distinct (List.sort is))

let test ()  =
    seq {
        for r1 in rects 5 5 do
          for r2 in rects 5 5 do
            for r3 in rects 5 5 do
              for r4 in rects 5 5 do
                for r5 in rects 5 5 do
                  yield (regions 5 5 [r1; r2; r3; r4; r5])
    }

(*

open Algorithmics.Area ;;
test () ;;
Seq.max it ;;
Seq.contains 5 it ;;
seq { for i in 17 .. 32 do yield Seq.contains i (test ()) } ;;
for i in 28 .. 32 do print (i, Seq.contains i (test ())) ;;       

*)
