(*

fsharpi Mini.fs Algorithmics/BinarySearch.fs

open Mini ;;
open Algorithmics.BinarySearch ;;

*)

module Algorithmics.BinarySearch

open Mini

let binarySearch (oracle : Nat -> Bool) (lower : Nat, upper : Nat) : Nat =
  let rec search (l, u) =
    if l >= u then u
    else  let  m = l + (u - l) / 2N
          in   if oracle m  then  search (l,       m)
                            else  search (m + 1N,  u)
  search (lower, upper)

// The following code demonstrates two problems with binary search
// (using signed bytes for demonstration purposes).

let flawedBinarySearch (oracle : sbyte -> Bool) (lower : sbyte, upper : sbyte) : sbyte =
  let rec search (l, u) =
    if l >= u then u
    else  let  m = (l + u) / 2y
          in   if oracle m  then  search (l,       m)
                            else  search (m + 1y,  u)
  search (lower, upper)

let box = [| 0; 1; 2; 3; 4 |]

// Infinite loop: F#'s integer division is T-division (truncating
// division, rounding towards zero) not F-division (floor division,
// rounding towards -∞).

let bug1 () = flawedBinarySearch (fun i -> 0 <= box.[int i + 2]) (-2y, 2y)

// Index out of range exception: l + u overflows.

let bug2 () = flawedBinarySearch (fun i -> 0 <= box.[int i - 100]) (100y, 104y)

let sbyteBinarySearch (oracle : sbyte -> Bool) (lower : sbyte, upper : sbyte) : sbyte =
  let rec search (l, u) =
    if l >= u then u
    else  let  m = l + (u - l) / 2y  // different calculation of mid-point
          in   if oracle m  then  search (l,       m)
                            else  search (m + 1y,  u)
  search (lower, upper)

let ok1 () = sbyteBinarySearch (fun i -> 0 <= box.[int i + 2]) (-2y, 2y)

let ok2 () = sbyteBinarySearch (fun i -> 0 <= box.[int i - 100]) (100y, 104y)

let intBinarySearch (oracle : int -> Bool) (lower : int, upper : int) : int =
  let rec search (l, u) =
    if l >= u then u
    else  let  m = l + (u - l) / 2
          in   if oracle m  then  search (l,      m)
                            else  search (m + 1,  u)
  search (lower, upper)
