(*

fsharpc -a Algorithmics/Map/Map.fsi Algorithmics/Map/List.fs

fsharpi -r List.dll
open Algorithmics ;;

An implementation of finite lists using unordered lists possibly
containing duplicates ie multiple entries with the same key.

*)

module Algorithmics.Map

// implementation of the interface

type Entry<'key, 'value> =
  { key : 'key; value : 'value }

type Map<'key, 'value when 'key : comparison> =
  | Rep of List<Entry<'key, 'value>>

let empty = Rep []

let add (key, value) (Rep list)
  = Rep ({ key = key; value = value } :: list)

let remove key (Rep list) =
  let rec del = function
  | []               -> []
  | entry :: entries -> if key = entry.key then del entries
                                           else entry :: del entries
  in  Rep (del list)

let isEmpty (Rep list) = List.isEmpty list

let lookup key (Rep list) =
  let rec find = function
  | []               -> None
  | entry :: entries -> if key = entry.key then Some entry.value
                                           else find entries
  in  find list

let fromList list =
  Rep (List.map (fun (key, value) -> { key = key; value = value }) list)

let toList (Rep list) =
  List.map (fun entry -> (entry.key, entry.value)) list

let toKVList (Rep list) = list

type Map<'key, 'value when 'key : comparison> with
  member this.Item
    with get key =
      match lookup key this with
      | None   -> raise (System.Collections.Generic.KeyNotFoundException())
      | Some v -> v
