(*

fsharpc -a Algorithmics/Map/Map.fsi Algorithmics/Map/OrdList.fs 

fsharpi -r OrdList.dll
open Algorithmics ;;

An implementation of finite maps using ordered list with no duplicates
ie no two entries have the same key.

*)

module Algorithmics.Map

// implementation of the interface

type Entry<'key, 'value> =
  { key : 'key; value : 'value }

type Map<'key, 'value when 'key : comparison> =
  | Rep of List<Entry<'key, 'value>>

let empty = Rep []

let add (key, value) (Rep list) =
  let newEntry = { key = key; value = value }
  let rec ins = function
  | []      -> [newEntry]
  | entry :: entries ->
      if   key < entry.key  then newEntry :: entry :: entries
      elif key = entry.key  then newEntry :: entries
      (* key > entry.key *) else entry :: ins entries
  in Rep (ins list)

let remove key (Rep list) =
  let rec del = function
  | []      -> []
  | x :: xs ->
      if   key < x.key  then  x :: xs
      elif key = x.key  then  xs
      (* key > x.key *) else x :: del xs
  in Rep (del list)

let isEmpty (Rep list) = List.isEmpty list

let lookup key (Rep list) =
  let rec find = function
  | []      -> None
  | x :: xs ->
      if   key < x.key  then  None
      elif key = x.key  then  Some x.value
      (* key > x.key *) else find xs
  in find list
        
let fromList list =
  Rep (List.map (fun (key, value) -> { key = key; value = value })
        (List.distinctBy fst 
          (List.sortBy fst list)))

let fromList2 list =
  Rep << List.map (fun (key, value) -> { key = key; value = value })
      << List.distinctBy fst
      << List.sortBy fst
      <| list

let fromList3 list =
  list |> List.sortBy fst
       |> List.distinctBy fst
       |> List.map (fun (key, value) -> { key = key; value = value })
       |> Rep

let toList (Rep list) =
  List.map (fun entry -> (entry.key, entry.value)) list

let toKVList (Rep list) = list

type Map<'key, 'value when 'key : comparison> with
  member this.Item
    with get key =
      match lookup key this with
      | None   -> raise (System.Collections.Generic.KeyNotFoundException())
      | Some v -> v
