(*

fsharpc -a Algorithmics/Map/Map.fsi Mini.fs Algorithmics/Tree.fs Algorithmics/Map/SearchTree.fs 
fsharpi -r SearchTree.dll

fsharpi Algorithmics/Map/Map.fsi Mini.fs Algorithmics/Tree.fs Algorithmics/Map/SearchTree.fs 

open Algorithmics.Map ;;

An implementation of finite maps using search trees with no
duplicates.

*)

module Algorithmics.Map

open Mini
open Algorithmics.Tree
open System.Collections.Generic

// implementation of the interface

type Entry<'key, 'value> =
  { key : 'key; value : 'value }

type Map<'key, 'value when 'key : comparison> =
  | Rep of Tree<Entry<'key, 'value>>

// to avoid duplication of code we use an extension method, see below
(*
  member this.Item
    with get key =
      let lookup key (Rep tree) =
        let rec find = function
          | Leaf           -> raise (KeyNotFoundException ())
          | Node (l, x, r) ->
              if   key < x.key  then find l
              elif key = x.key  then x.value
              (* key > x.key *) else find r
        in find tree
      in lookup key this
*)

let empty = Rep Leaf

let add (key, value) (Rep tree) =
  let newEntry = { key = key; value = value }
  let rec ins = function
  | Leaf           -> Node (Leaf, newEntry, Leaf)
  | Node (l, x, r) ->
      if   key < x.key  then Node (ins l, x, r)
      elif key = x.key  then Node (l, newEntry, r)
      (* key > x.key *) else Node (l, x, ins r)
  in Rep (ins tree)

let remove key (Rep tree) =
  let rec del = function
  | Leaf           -> Leaf
  | Node (l, x, r) ->
      if   key < x.key  then Node (del l, x, r)
      elif key = x.key  then join l r
      (* key > x.key *) else Node (l, x, del r)
  in Rep (del tree)

let isEmpty (Rep tree) =
  match tree with
  | Leaf           -> true
  | Node (_, _, _) -> false

let lookup key (Rep tree) =
  let rec find = function
  | Leaf           -> None
  | Node (l, x, r) ->
      if   key < x.key  then find l
      elif key = x.key  then Some x.value
      (* key > x.key *) else find r
  in find tree

let fromList list =
  list |> List.sortBy     fst
       |> List.distinctBy fst
       |> List.map        (fun (key, value) -> { key = key; value = value })
       |> balancedTree
       |> Rep

let toList (Rep tree) =
  List.map (fun entry -> (entry.key, entry.value)) (inorder tree)

type Map<'key, 'value when 'key : comparison> with
  member this.Item
    with get key =
      match lookup key this with
      | None   -> raise (KeyNotFoundException())
      | Some v -> v

// let m = add (4711, "Anja") empty ;;
// m.[4711];;
