(*

fsharpi Mini.fs Knobelaufgaben/Sort.fs Algorithmics/Median.fs

open Algorithmics.Median ;;

*)

module Algorithmics.Median

open Knobelaufgaben.Sort

// Ordnungsstatistik und Median.

let MEdian3 (a, b, c) =
    let (_, m, _) = sort3 (a, b, c) in m
let MEdian5 (a, b, c, d, e) =
    let (_, _, m, _, _) = sort5 (a, b, c, d, e) in m

let (</) a b  =  if a <= b then a else b
let (/>) a b  =  if a <= b then b else a

//let Median3 (a : 'a, b : 'a, c : 'a) : 'a when 'a : comparison =
let Median3 (a, b, c) =
    (a </ b) /> (a </ c) /> (b </ c)

let Median4 (a, b, c, d) =
    (a </ b </ c) /> (a </ b </ d) /> (a </ c </ d) /> (b </ c </ d)

let Median5 (a, b, c, d, e) =
    (a </ b </ c) /> (a </ b </ d) /> (a </ b </ e) /> (a </ c </ d) /> (a </ c </ e) />
    (a </ d </ e) /> (b </ c </ d) /> (b </ c </ e) /> (b </ d </ e) /> (c </ d </ e)

//let MEdian3 (a, b, c) =
//    let (l, g) = sort2 (a, b)
//    hmmm
    
let median3 (a, b, c) =
    (a </ (b /> c)) /> (b </ c)

let median4 (a, b, c, d) =
    (a /> b) </ (c /> d) </ ((a </ b) /> (c </ d))
// { a /> b, c /> d, (a </ b) /> (c </ d) } ist die Menge ohne das kleinste Element
   
// https://stackoverflow.com/questions/480960/code-to-calculate-median-of-five-in-c-sharp

let median5 (a, b, c, d, e) =
    let (l1, g1) = sort2 (a, b)
    let (l2, g2) = sort2 (c, d)
// { l1 /> l2, g1 </ g2 } ist die Menge der beiden mittleren Elemente.    
    Median3 (l1 /> l2, g1 </ g2, e)

let xs = [0 .. 9]
// List.reduce (&&) [for i in 0 .. 9 do for j in 0 .. 9 do for k in 0 .. 9 do yield (median3 (i,j,k) = Median3 (i,j,k)) ] ;;
// List.reduce (&&) [for a in xs do for b in xs do for c in xs do for d in xs do for e in xs do yield (median5 (a,b,c,d,e) = MEdian5 (a,b,c,d,e)) ] ;;

// Hoare's quick select

let rec filter (good : 'a -> Bool) : 'a list -> 'a list = function
  | []       ->  []
  | x :: xs  ->  if good x  then  x ::  filter good xs 
                            else        filter good xs

let oracle xs =
    List.head xs

let rec quickselect k xs =
    let pivot = oracle xs
    let lts   = List.filter (fun x -> x < pivot) xs
    let eqs   = List.filter (fun x -> x = pivot) xs
    let gts   = List.filter (fun x -> x > pivot) xs
    let i     = List.length lts
    let j     = i + List.length eqs
    
    if   k <= i then quickselect k lts
    elif k <= j then pivot
                else quickselect (k - j) gts

// Blum et al median of median
// https://en.wikipedia.org/wiki/Median_of_medians

let rec mediansOf5 = function
    | []                                -> []
    | [x]                               -> [x]
    | [x1; x2]                          -> [ x1 </ x2 ] 
    | [x1; x2; x3]                      -> [median3 (x1, x2, x3)]
    | [x1; x2; x3; x4]                  -> [median4 (x1, x2, x3, x4)]
    | x1 :: x2 :: x3 :: x4 :: x5 :: xs  -> median5 (x1, x2, x3, x4, x5) :: mediansOf5 xs

let rec median = function
    | [x]                   -> x
    | [x1; x2]              -> x1 </ x2
    | [x1; x2; x3]          -> median3 (x1, x2, x3)
    | [x1; x2; x3; x4]      -> median4 (x1, x2, x3, x4)
    | [x1; x2; x3; x4; x5]  -> median5 (x1, x2, x3, x4, x5)
    | xs                    -> ordselect ((List.length xs - 1) / 2) xs

and ordselect k xs =
    let pivot = median (mediansOf5 xs)
    let lts   = List.filter (fun x -> x < pivot) xs
    let eqs   = List.filter (fun x -> x = pivot) xs
    let gts   = List.filter (fun x -> x > pivot) xs
    let i     = List.length lts
    let j     = i + List.length eqs
    
    if   k <= i then ordselect k lts
    elif k <= j then pivot
                else ordselect (k - j) gts

module NEW =
    let rec quickselect (oracle : 'a list -> 'a) i xs =
        let pivot = oracle xs
        
        let lts   = List.filter (fun x -> x < pivot) xs
        let eqs   = List.filter (fun x -> x = pivot) xs
        let gts   = List.filter (fun x -> x > pivot) xs
        let k     = List.length lts
        let l     = k + List.length eqs

        if   i <= k then quickselect oracle i lts
        elif i <= l then pivot
                    else quickselect oracle (i - l) gts

    let rec ordselect i xs =
        quickselect (median << mediansOf5) i xs
        
    and median = function
        | [x]                   -> x
        | [x1; x2]              -> x1 </ x2
        | [x1; x2; x3]          -> median3 (x1, x2, x3)
        | [x1; x2; x3; x4]      -> median4 (x1, x2, x3, x4)
        | [x1; x2; x3; x4; x5]  -> median5 (x1, x2, x3, x4, x5)
        | xs                    -> ordselect (List.length xs / 2) xs

let main =
    print (NEW.ordselect 5000 [1..9999])

// [ for i in 1 .. 999 -> NEW.ordselect i [1 .. 999] ] = [1 .. 999] ;;
// [ for i in 1 .. 999 -> NEW.ordselect i [999 .. -1 .. 1] ] = [1 .. 999] ;;    
