// https://stackoverflow.com/questions/36847449/how-to-write-code-in-f-for-what-functors-do-in-ocaml

(*

fsi Algorithmics/PQueue/PQueue.fsi Algorithmics/PQueue/Pennant.fs Algorithmics/PQueue/BinomialHeap.fs
open Algorithmics.PQueue ;;

An implementation of priority queues using binomial heaps.

*)

module Algorithmics.PQueue

// implementation of the interface

open Algorithmics.Pennant

type Bit<'a> =
    | Zero
    | One  of Pennant<'a>

type PQueue<'a> =
    | Rep of List<Bit<'a>>

let rec size = function
    | []          -> 0
    | Zero :: ps  -> 0 + 2 * size ps
    | One _ :: ps -> 1 + 2 * size ps

let empty = Rep []

let single a = Rep [One (a, Leaf)]

let rec succ = function
    | (p, [])          -> [One p]
    | (p, Zero :: ps)  -> One p :: ps
    | (p, One q :: ps) -> Zero :: succ (versus (p, q), ps)

let insert (a, Rep q) = Rep (succ ((a, Leaf), q))

let rec add = function
    |  ([], ps) | (ps, [])        -> ps
    |  (Zero :: ps, Zero :: qs)   -> Zero :: add (ps, qs)
    |  (Zero :: ps, One p :: qs)
    |  (One p :: ps, Zero :: qs)  -> One p :: add (ps, qs)
    |  (One p :: ps, One q :: qs) -> Zero :: succ (versus (p, q), add (ps, qs))

let meld (Rep q1, Rep q2) = Rep (add (q1, q2))

(* remove minimum pennant and return remainder *)

let rec splitMinPennant = function
    | []         -> None
    | Zero :: ps ->  match splitMinPennant ps with
                     | None         -> None 
                     | Some (q, []) -> Some (q, [])  (* avoid trailing 0s *)
                     | Some (q, qs) -> Some (q, Zero :: qs)
    | One p :: ps -> match splitMinPennant ps with
                     | None         -> Some (p, [])
                     | Some (q, qs) -> if fst p <= fst q then Some (p, Zero :: ps)
                                                         else Some (q, One p :: qs)

let splitMin (Rep ps) =
    match splitMinPennant ps with
    | None              -> None
    | Some ((a, t), qs) -> Some (a, Rep (add (List.map One (List.rev (spine t)), qs)))

///////////////////////////////////////////////////////////////////////////////
// Default.skel

let rec toOrdList q =
    match splitMin q with
    | None         -> []
    | Some (x, q') -> x :: toOrdList q'

let rec playRound = function
    | []             -> []
    | [q1]           -> [q1]
    | q1 :: q2 :: qs -> meld (q1, q2) :: playRound qs

let rec tournament = function
    | []  -> empty
    | [q] -> q
    | qs  -> tournament (playRound qs)

let bottomUp (xs : 'elem list) =
    tournament [ for x in xs -> single x ]

(*
let rec rightToLeft = function
    | []      -> empty
    | x :: xs -> meld (single x, rightToLeft xs)
//    | x :: xs -> insert (x, rightToLeft xs)

let rec rightToLeft = function
    | []      -> empty
    | q :: qs -> meld (q, rightToLeft qs)
*)

let rightToLeft (xs : 'elem list) =
    List.foldBack (fun a q -> insert (a, q)) xs empty

let fromList xs = rightToLeft xs

let sort xs = toOrdList (fromList xs)

(*
let xs = [1..100000] ;;     
let xs = [100000 .. -1 .. 1] ;;
let rnd = System.Random(4711) ;;
let xs = List.init 100000 (fun _ -> rnd.Next ()) ;;
#time ;;
sort xs ;;
*)   

///////////////////////////////////////////////////////////////////////////////
