(*

An implementation of priority queues using unordered lists possibly
containing duplicates.

fsi PQueue.fsi List.fs
open Algorithmics.PQueue ;;

*)

module Algorithmics.PQueue

// implementation of the interface

type PQueue<'elem when 'elem : comparison> =
  | Rep of List<'elem>

let empty = Rep []

let single x = Rep [x]

let meld (Rep q1, Rep q2) = Rep (q1 @ q2)

let insert (x, q) = meld (single x, q)

let splitMin (Rep q) =
    let rec split = function
        | []      -> None
        | x :: xs -> Some (match split xs with
                           | None         -> (x, xs)  // xs = []
                           | Some (m, ys) -> if x <= m then (x, xs)
                                             else (m, x :: ys))
    match split q with
    | None        -> None
    | Some (a, q) -> Some (a, Rep q)

///////////////////////////////////////////////////////////////////////////////
// Default.skel

(* bulk operations *)

let rec toOrdList q =
    match splitMin q with
    | None         -> []
    | Some (x, q') -> x :: toOrdList q'

let rec playRound = function
    | []             -> []
    | [q1]           -> [q1]
    | q1 :: q2 :: qs -> meld (q1, q2) :: playRound qs

let rec tournament = function
    | []  -> empty
    | [q] -> q
    | qs  -> tournament (playRound qs)

let bottomUp (xs : 'elem list) =
    tournament [ for x in xs -> single x ]

let rec rightToLeft = function
    | []      -> empty
    | x :: xs -> meld (single x, rightToLeft xs)

(*
let rec rightToLeft = function
    | []      -> empty
    | q :: qs -> meld (q, rightToLeft qs)
*)

let fromList xs = Rep xs

let sort xs = toOrdList (fromList xs)

(*
let xs = [1..100000] ;;     
let xs = [100000 .. -1 .. 1] ;;
let rnd = System.Random(4711) ;;
let xs = List.init 100000 (fun _ -> rnd.Next ()) ;;
#time ;;
sort xs ;;
*)   

///////////////////////////////////////////////////////////////////////////////
