(*

An implementation of priority queues using ordered lists possibly
containing duplicates.

fsi Algorithmics/PQueue/PQueue.fsi Algorithmics/PQueue/OrdList.fs
open Algorithmics.PQueue ;;

*)

module Algorithmics.PQueue

// implementation of the interface

type PQueue<'elem when 'elem : comparison> =
  | Rep of List<'elem>

let empty = Rep []

let single x = Rep [x]

let rec merge = function
  | ([], xs) | (xs, []) -> xs
  | (x :: xs, y :: ys)  -> if x <= y then x :: merge (xs, y :: ys)
                                     else y :: merge (x :: xs, ys)

let meld (Rep q1, Rep q2) = Rep (merge (q1, q2))

let rec ins k = function
    | []      -> [k]
    | x :: xs -> if k <= x then k :: x :: xs
                           else x :: ins k xs

let insert (x, Rep q) = Rep (ins x q)

let splitMin (Rep q) =
    let rec split = function
        | []      -> None
        | x :: xs -> Some (match split xs with
                           | None         -> (x, xs)  // xs = []
                           | Some (m, ys) -> if x <= m then (x, xs)
                                             else (m, x :: ys))
    match split q with
    | None        -> None
    | Some (a, q) -> Some (a, Rep q)

///////////////////////////////////////////////////////////////////////////////
// Default.skel

(* bulk operations *)

let toOrdList (Rep q) = q

let rec playRound = function
    | []             -> []
    | [q1]           -> [q1]
    | q1 :: q2 :: qs -> meld (q1, q2) :: playRound qs

let rec tournament = function
    | []  -> empty
    | [q] -> q
    | qs  -> tournament (playRound qs)

let bottomUp (xs : 'elem list) =
    tournament [ for x in xs -> single x ]

(*
let rec rightToLeft = function
    | []      -> empty
    | x :: xs -> meld (single x, rightToLeft xs)

let rec rightToLeft = function
    | []      -> empty
    | q :: qs -> meld (q, rightToLeft qs)
*)

let rightToLeft (xs : 'elem list) =
    List.foldBack (fun a q -> insert (a, q)) xs empty

let fromList xs = bottomUp xs

let sort xs = toOrdList (fromList xs)

(*
let xs = [1..100000] ;;     
let xs = [100000 .. -1 .. 1] ;;
let random = System.Random(4711) ;;
let xs = [ for i in 1 .. 100000 -> random.Next () ] ;;
#time ;;
sort xs ;;
*)   

///////////////////////////////////////////////////////////////////////////////
