(*

An implementation of priority queues using topped looser trees /
pairing heaps (they differ only in the code for splitMin).

fsi Algorithmics/PQueue/PQueue.fsi Algorithmics/PQueue/Pennant.fs Algorithmics/PQueue/Tournament.fs
open Algorithmics.PQueue ;;

dotnet new classlib -lang "F#"
dotnet build
fsi -r bin/Debug/netstandard2.0/PQueue.dll

*)

module Algorithmics.PQueue
#nowarn "62"

open Algorithmics.Pennant

// implementation of the interface

// same as ('elem * Tree<'elem>) option
type PQueue<'elem when 'elem : comparison> =
  | Inf
  | Min of Pennant<'elem>

let empty = Inf

let single x = Min (x, Leaf)

let meld = function
    | (Inf, q) | (q, Inf)       -> q
    | (Min (a, t), Min (b, u))  -> Min (versus ((a, t), (b, u)))

let insert (x, q) = meld (single x, q)

let rec secondBest1 = function
    | Leaf           -> Inf
    | Node (t, a, u) -> meld (Min (a, t), secondBest1 u)

let rec secondBest2 = function
    | Leaf                        -> Inf
    | Node (t, a, Leaf)           -> Min (a, t)
    | Node (t, a, Node (u, b, v)) -> meld (Min (versus ((a, t), (b, u))), secondBest2 v)
//    | Node (t, a, Node (u, b, v)) -> meld (meld (Min (a, t), Min (b, u)), secondBest2 v)

let splitMin = function
    | Inf        -> None
    | Min (a, t) -> Some (a, secondBest2 t)

///////////////////////////////////////////////////////////////////////////////
// Default.skel

let rec toOrdList q =
    match splitMin q with
    | None         -> []
    | Some (x, q') -> x :: toOrdList q'

let rec playRound = function
    | []             -> []
    | [q1]           -> [q1]
    | q1 :: q2 :: qs -> meld (q1, q2) :: playRound qs

let rec tournament = function
    | []  -> empty
    | [q] -> q
    | qs  -> tournament (playRound qs)

let bottomUp (xs : 'elem list) =
    tournament [ for x in xs -> single x ]

(*
let rec rightToLeft = function
    | []      -> empty
    | x :: xs -> meld (single x, rightToLeft xs)

let rec rightToLeft = function
    | []      -> empty
    | q :: qs -> meld (q, rightToLeft qs)
*)

//let fromList xs = bottomUp xs
//let fromList xs = rightToLeft xs
let rightToLeft (xs : 'elem list) =
    List.foldBack (fun a q -> insert (a, q)) xs empty

let fromList xs = bottomUp xs

let sort xs = toOrdList (fromList xs)

(*
let xs = [1..100000] ;;     
let xs = [100000 .. -1 .. 1] ;;
let rnd = System.Random(4711) ;;
let xs = List.init 100000 (fun _ -> rnd.Next ()) ;;
let xs = [ for i in 1 .. 100000 -> rnd.Next () ] ;;
#time ;;
sort xs ;;
*)   
