(*

fsharpi Mini.fs Datatypes/Person.fs Algorithmics/Tree.fs Algorithmics/BinarySearch.fs Algorithmics/Search.fs

open Mini ;;
open Algorithmics.Search ;;

*)

module Algorithmics.Search

open Mini
open Datatypes.Person

type Entry = { key : Nat; person : Person }

let team =  [ { key = 7N;     person = ralf    };
              { key = 815N;   person = melanie };
              { key = 4711N;  person = julia   };
              { key = 4712N;  person = andres  } ]

// Suchen in einer ungeordneten Liste.

module Unordered =
  let rec lookup (key : Nat, staff : List<Entry>) : Option<Person> =
    match staff with
    | []               -> None
    | entry :: entries ->
        if key = entry.key then Some entry.person
                           else lookup (key, entries)

// Suchen in einer geordneten Liste.

module Ordered =
  let rec lookup (key : Nat, staff : List<Entry>) : Option<Person> =
    match staff with
    | []               ->  None
    | entry :: entries ->
        if   key < entry.key  then None
        elif key = entry.key  then Some entry.person
        (* key > entry.key *) else lookup (key, entries)

// Suchen in einem geordneten Baum, einem Suchbaum.
         
module SearchTree =
  open Algorithmics.Tree
  
  let rec lookup (key : Nat, staff : Tree<Entry>) : Option<Person> =
    match staff with
    | Empty                     -> None
    | Node (left, entry, right) ->
        if   key < entry.key  then lookup (key, left)
        elif key = entry.key  then Some entry.person
        (* key > entry.key *) else lookup (key, right)

open Algorithmics.BinarySearch

module OrderedArray =
  let team =  [| { key = 7N;     person = ralf    };
                 { key = 815N;   person = melanie };
                 { key = 4711N;  person = julia   };
                 { key = 4712N;  person = andres  } |]

  let lookup (key : Nat, staff : Array<Entry>) : Option<Person> =
    let i = intBinarySearch (fun k -> key <= staff.[k].key) (0, staff.Length - 1)
    if key = staff.[i].key then Some staff.[i].person else None
