(*

fsharpc -a Algorithmics/Set/Set.fsi Mini.fs Algorithmics/Tree.fs Algorithmics/Set/SearchTree.fs

fsharpi -r SearchTree.dll
open Algorithmics ;;

An implementation of sets using search trees with no duplicates.

*)

module Algorithmics.Set

open Mini
open Algorithmics.Tree

// implementation of the interface

type Set<'elem when 'elem : comparison> =
  | Abs of Tree<'elem>

let empty = Abs Leaf

let add (key, Abs tree) =
  let rec ins = function
  | Leaf           -> Node (Leaf, key, Leaf)
  | Node (l, x, r) ->
      if   key < x  then Node (ins l, x, r)
      elif key = x  then Node (l, key, r)
      else (* key > x *) Node (l, x, ins r)
  in Abs (ins tree)

let remove (key, Abs tree) =
  let rec del = function
  | Leaf           -> Leaf
  | Node (l, x, r) ->
      if   key < x  then Node (del l, x, r)
      elif key = x  then join l r
      else (* key > x *) Node (l, x, del r)
  in Abs (del tree)

let isEmpty (Abs tree) =
  match tree with
  | Leaf           -> true
  | Node (_, _, _) -> false

let contains (key, Abs tree) =
  let rec find = function
  | Leaf           -> false
  | Node (l, x, r) ->
      if   key < x  then find l
      elif key = x  then true
      else (* key > x *) find r
  in find tree

let fromList list = Abs (balancedTree (List.distinct (List.sort list)))

let toList (Abs tree) = inorder tree
