(*
Bank accounts.

fsharpc Mini.fs Effects/Bank.fs
mono Bank.exe

fsharpi Mini.fs Effects/Bank.fs
open Effects.Bank ;;
*)

module Effects.Bank
#nowarn "62"

open Mini

[<RequireQualifiedAccess>]
module Account =
  let private funds = ref 0N

  let deposit (amount : Nat) =
      funds := !funds + amount

  let withdraw (amount : Nat) =
      let old = !funds
      funds := !funds - amount
      old - !funds

  let balance () = !funds

exception Insufficient of Nat

[<RequireQualifiedAccess>]
module JuniorAccount =

  let private minFunds = 10N
  let private funds = ref 0N

  let deposit (amount : Nat) =
      funds := !funds + amount

  let withdraw (amount : Nat) =
    if !funds - amount >= minFunds then
      funds := !funds - amount
    else
      raise (Insufficient (!funds - minFunds))

  let balance () =
      !funds

let maximal_withdraw (amount : Nat) : Nat =
  try
    JuniorAccount.withdraw amount; amount
  with
    Insufficient rest ->
      JuniorAccount.withdraw rest; rest

// JuniorAccount.deposit 4711N ;;
// JuniorAccount.withdraw 815N ;;
// JuniorAccount.withdraw 2765N ;;
// JuniorAccount.withdraw 2765N ;;

let main =
    JuniorAccount.deposit 4711N
    JuniorAccount.withdraw 0815N
    putline ("funds = " ^ show (JuniorAccount.balance ()))
    JuniorAccount.withdraw 2765N
    putline ("funds = " ^ show (JuniorAccount.balance ()))
    putline ("withdrawn = " ^ show (maximal_withdraw 2765N))
    putline ("funds = " ^ show (JuniorAccount.balance ()))
