(*
Control structures: for and while loops.

fsharpi Mini.fs Effects/Control.fs
open Effects.Control ;;
*)

module Effects.Control

open Mini

let rec foreach (lower : Nat, upper : Nat) (body : Nat -> Unit) : Unit =
  if lower <= upper then
    body lower
    foreach (lower + 1N, upper) body

let rec foreach' (lower : Nat, upper : Nat) (body : Nat -> Unit) : Unit =
  if lower > upper  then  ()
                    else  body lower; 
                          foreach' (lower + 1N, upper) body

// foreach (0N, 9N) (fun i -> print (i * i)) ;;
// for i in 0N .. 9N do print (i * i) ;;

let nat (i : int) = Nat.Make i

let rec foreachArray (array : Array<'a>) (body : 'a -> Unit) =
  foreach (0N, nat array.Length - 1N)
    (fun i -> body (array.[int i]))

let rec foreachList (list : List<'a>) (body : 'a -> Unit) =
  match list with
    | []       ->  ()
    | x :: xs  ->  body x ; foreachList xs body

// foreachArray [|0N .. 9N|] (fun i -> print (i * i)) ;;
// foreachList [0N .. 9N] (fun i -> print (i * i)) ;;
// for i in [|0N .. 9N|] do print (i * i) ;;
// for i in [0N .. 9N] do print (i * i) ;;

type Block<'value> = Unit -> 'value

let rec whiledo (test : Block<Bool>) (body : Block<Unit>) : Unit =
  if test () then
    body ()
    whiledo test body

let sum (a : Array<Int>) =
    let mutable i = 0
    let mutable s = 0
    whiledo (fun () -> i < a.Length)
            (fun () -> s <- s + a.[i]
                       i <- i + 1)
    s

// sum [| 1.. 100 |] ;;
