(*
Database queries using list comprehensions.

fsharpi Mini.fs Effects/Database.fs
open Effects ;;
*)

[<AutoOpen>]
module Effects.Database

open Mini

// Consider the following types and signatures for a simple database
// of students and exam results.

type Name    =  String   // student name
type GUID    =  Int      // student id (Globally Unique Identifier)

type Course  =  | ALG | FPR | IPR | OOP
type Mark    =  | Insufficient | Sufficient | Good | VeryGood | Excellent

type Student =  { name : Name; guid : GUID }
type Result  =  { course : Course; guid : GUID; mark : Mark }

let students : Student list =
    [{name = "Ralf"; guid = 4711};
     {name = "Lisa"; guid = 815}]
let results : Result list =
    [{course = ALG; guid = 4711; mark = Good};
     {course = FPR; guid = 815;  mark = Excellent};
     {course = FPR; guid = 4711; mark = Sufficient}]

// What is the name of the student with student number 4711?

let query1 : Name =
    List.head [ for student in students do
                  if student.guid = 4711 then
                    yield student.name ]

// What are the results of the course ALG? Return a list of marks.

let query2 : Mark list =
    [ for result in results do
        if result.course = ALG then
          yield result.mark ]

// Who scored highest in the course FPR? It is sufficient to return
// a student number.

let query3 : GUID =
    snd (List.max [ for result in results do
                      if result.course = FPR then
                         yield (result.mark, result.guid) ])

// What are the results of the student(s) named "Ralf"? Return a
// list of marks.

let query4 : Mark list =
    [ for student in students do
        if student.name = "Ralf" then
          for result in results do
            if student.guid = result.guid then
              yield result.mark ]

// Which students had at least one excellent result in one of their
// courses? Return a list of names.

let query5 : Name list =
    [ for result in results do
        if result.mark = Excellent then
          for student in students do
            if result.guid = student.guid then
              yield student.name ]
