(*
Ephemeral lists (tail is a reference cell).

fsharpc Mini.fs Effects/ImperativeList.fs
mono ImperativeList.exe

fsharpi Mini.fs Effects/ImperativeList.fs
open Effects.ImperativeList ;;
*)

module Effects.ImperativeList

open Mini

type List<'elem> =
    Ref<Item<'elem>>
 and Item<'elem> =
    | Nil
    | Cons of 'elem * List<'elem>

let nil () = ref Nil
let cons (x, xs) = ref (Cons (x, xs))

let head (list : List<'elem>) : 'elem =
  match !list with
    | Cons (x, xs) -> x

let tail (list : List<'elem>) : List<'elem> =
  match !list with
    | Cons (x, xs) -> xs

let rec append (list1 : List<'elem>, list2 : List<'elem>) =
  match !list1 with
    | Nil          -> list1 := !list2
    | Cons (x, xs) -> append (xs, list2)

// tail recursive
let length (list : List<'elem>) =
  let rec worker n p = 
    match !p with
      | Nil          -> n
      | Cons (x, xs) -> worker (n + 1) xs
  in worker 0 list

let length' (list : List<'elem>) =
  let mutable n = 0
  let mutable cur = list
  while !cur <> Nil do
      n <- n + 1
      cur <- tail cur
  n

let primes = cons (2, cons (3, cons (5, cons (7, cons (11, nil ())))))

// length primes ;;
// append (primes, primes) ;;
// length primes ;;
