(*
Eingabe mit Validierung (Option).

fsharpc Mini.fs Datatypes/Person.fs Effects/Input.fs
mono Input.exe

fsharpi Mini.fs Datatypes/Person.fs Effects/Input.fs
open Effects.Input ;;
*)

module Effects.Input
#nowarn "62"

open Mini

let query (prompt : String) : String =
    putstring prompt; getline ()

let humanPlayer (guess : Nat) : Bool =
    query ("Ist die Zahl gleich oder kleiner als " ^ show guess ^ "? ") = "ja"

let rec factorial (n : Nat) : Nat =
    putstring (show n ^ "\n")
    if n = 0N then 1N else factorial (n - 1N) * n

// ohne Plausibilitätsüberprüfung

open Datatypes.Person

let simpleInputPerson () : Person =
  if List.contains (query "gender: ") ["f"; "female"] then
    Female  {  name  =  query "name:   " }
  else
    Male    {  name  =  query "name:   ";
               bald  =  List.contains (query "bald?:  ") ["y"; "yes"] }

// mit Plausibilitätsüberprüfung

type Result<'a> =
    | Okay  of 'a
    | Error of String

let rec checkedQuery (prompt : String, check : String -> Result<'a>) : 'a =
  match check (query (prompt ^ ": ")) with
  | Okay  v   -> v
  | Error msg -> putline ("*** " ^ msg) ;
                 checkedQuery (prompt, check)

// String.forall (fun x -> x = 'a') "aaa" ;;
// System.Char.IsDigit 'a' ;;

open System

let isNat (s : String) : Result<Nat> =
  if s <> "" && String.forall Char.IsDigit s then
    Okay (Nat.Parse s)
  else
    Error "natural number expected"

// checkedQuery ("age", isNat) ;;

// Komposition von Validatoren.
let both (first: 'a -> Result<'b>, second: 'b -> Result<'c>) : 'a -> Result<'c> =
  fun x -> match first x with
           | Okay  y   -> second y
           | Error msg -> Error msg

let isLess (n : Nat) : Nat -> Result<Nat> = fun m ->
    if m < n then Okay m
             else Error ("number must be less than " ^ n.ToString ())
let isGreater (n : Nat) : Nat -> Result<Nat> =
  fun m -> if m > n then Okay m
                    else Error ("number must be greater than " ^ n.ToString ())

// checkedQuery ("age", both (isNat, isLess 123N)) ;;
// checkedQuery ("age", both (both (isNat, isLess 123N), isGreater 17N)) ;;

let choice (prompt : String, choices : List<String>) : String =
  checkedQuery (prompt, 
    fun s -> 
      if List.contains s choices
      then Okay s
      else Error ("choices: " ^ String.concat ", " choices))

// choice ("gender", ["f"; "m"; "female"; "male"]) ;;

let isName = Okay  // too difficult to implement

let inputPerson () : Person =
  if List.contains (choice ("gender", ["f"; "m"; "female"; "male"])) ["f"; "female"] then
    Female { name = checkedQuery ("name  ", isName) }
  else
    Male   { name = checkedQuery ("name  ", isName);
             bald = List.contains (choice ("bald? ", ["y"; "n"; "yes"; "no"])) ["y"; "yes"] }

// print (inputPerson ())
