(*
Re-Implementierung der Eingabe mit Validierung mit Hilfe von
Ausnahmen.

fsharpc Mini.fs Datatypes/Person.fs Effects/Input.fs Effects/Input2.fs
mono Input2.exe

fsharpi Mini.fs Datatypes/Person.fs Effects/Input.fs Effects/Input2.fs
open Effects.Input2 ;;
*)

module Effects.Input2
#nowarn "62"

open Mini
open Datatypes.Person
open Effects

let rec checkedQuery (prompt : String, check : String -> 'a) : 'a =
  try
    check (Input.query (prompt ^ ": "))
  with
    Panic msg ->
      putline ("*** " ^ msg); checkedQuery (prompt, check)

exception Read

let readnat (s : String) : Nat =
  let rec nat = function
      | []      -> 0N
      | c :: cs -> if Char.IsDigit c then
                     Nat.ord c - Nat.ord '0' + 10N * nat cs
                   else
                     raise Read
  in
    if s = "" then raise Read else nat (List.rev (explode s))

let isNat (s : String) : Nat =
  try readnat s with Read -> panic "natural number expected"

// checked-read-line ("age", nat?) ;;

// Komposition von Validatoren
let both (first: 'a -> 'b, second: 'b -> 'c) : 'a -> 'c =
  fun x -> second (first x)

let isLess (n : Nat) : Nat -> Nat =
  fun m -> if m < n then m
                    else panic ("number must be less than " ^ show n)
let isGreater (n : Nat) : Nat -> Nat =
  fun m -> if m > n then m
                    else panic ("number must be greater than " ^ show n)

// checkedQuery ("age", both (isNat, isLess 123N)) ;;
// checkedQuery ("age", both (both (isNat, isLess 123N), isGreater 17N)) ;;

let choice (prompt : String, choices : List<String>) : String =
  checkedQuery (prompt, 
    fun s -> 
      if List.contains s choices
      then s
      else panic ("choices: " ^ String.concat ", " choices))

// choice ("gender", ["f"; "m"; "female"; "male"]) ;;

let isName = id  // too difficult to implement

let inputPerson () : Person =
  if List.contains (choice ("gender", ["f"; "m"; "female"; "male"])) ["f"; "female"] then
    Female { name = checkedQuery ("name  ", isName) }
  else
    Male   { name = checkedQuery ("name  ", isName);
             bald = List.contains (choice ("bald? ", ["y"; "n"; "yes"; "no"])) ["y"; "yes"] }

// inputPerson () ;;
