(*

Memoization: immediate/on-demand, excluding/including recursive calls.

fsharpc Mini.fs Effects/Memoization.fs
mono Memoization.exe

fsharpi Mini.fs Effects/Memoization.fs
open Effects.Memoization ;;

*)

module Effects.Memoization

open Mini

// cheating (arrays are indexed by ints); don't compute fibonacci 50
type Nat = int

let rec fibonacci (n : Nat) : Nat =
  if n <= 1 then n
            else fibonacci (n - 1) + fibonacci (n - 2)

// Memoisierung einer (nicht-rekursiven) Funktion.

let memo1 (dom : Nat, func : Nat -> 'v) : Nat -> 'v =
  let
//    memoTable = Array.init dom func
    memoTable = [| for i in 0 .. dom - 1 -> func i |]
  in
    fun (n : Nat) -> memoTable.[n]

let memoFibonacci1 = memo1 (10, fibonacci)

// Bedarfsgetriebene Memoisierung einer (nicht-rekursiven) Funktion.

let memo2 (dom : Nat, func : Nat -> 'v) : Nat -> 'v =
  let
//    memoTable = Array.create dom None
    memoTable = [| for i in 0 .. dom - 1 -> None |]
  in
    fun (n : Nat) ->
      match memoTable.[n] with
      | None   -> let v = func n
                  do memoTable.[n] <- Some v
                  v
      | Some v -> v

let memoFibonacci2 = memo2 (10, fibonacci)

// Fixpunkt einer Funktion mit Memoisierung.

let recMemo (dom : Nat, functional : (Nat -> 'v) -> (Nat -> 'v)) : Nat -> 'v =
//  let memoTable = Array.create dom None
  let memoTable = [| for i in 0 .. dom - 1 -> ref None |]
  
  let rec memoF (n : Nat) : 'v =
    match !memoTable.[n] with
    | None   -> let v = functional memoF n
                do memoTable.[n] := Some v
                v
    | Some v -> v
  in
    memoF

let memoFibonacci =
  recMemo (100,
    fun fib -> fun n ->
      if n <= 1 then n
                else fib (n - 1) + fib (n - 2))
