(*
Ephemeral implementation of non-empty sequences using necklaces, aka
cyclic lists.

fsharpc Mini.fs Effects/Necklace.fs
mono Necklace.exe

fsharpi Mini.fs Effects/Necklace.fs
open Effects.Necklace ;;
*)

module Effects.Necklace

open Mini

[<ReferenceEquality>]
type Necklace<'elem> =
    { mutable bead : 'elem
      mutable next : Necklace<'elem> }

let single x =
    let rec item =
        { bead = x
          next = item }
    in item

// swap can be used for catenation and for splitting
let swap (p : Necklace<'elem>, q : Necklace<'elem>) =
    let tmp = p.next
    p.next <- q.next
    q.next <- tmp

let append (last1 : Necklace<'elem>, last2 : Necklace<'elem>) : Necklace<'elem> =
    swap (last1, last2)
    last2

let cons (x, xs) = append (single x, xs)

let head (last : Necklace<'elem>) : 'elem =
    last.next.bead

let tail (last : Necklace<'elem>) : Necklace<'elem> =
//    last.next <- last.next.next
    swap (last, last.next)
    last

// let revenue = cons (4711, single 815) ;;
// let mutable primes = cons (2, single 3) ;;
// let more = cons (5, cons (7, single 11)) ;;
// primes <- append (primes, more) ;;

let rec replicate (n : Nat) (a :'a) : Necklace<'a> =  // a necklace with n + 1 beads
    if n = 0N then single a
              else append (replicate (n - 1N) a, single a)

// recognizes the language (a | b)^* b (a | b)^n
let accept (n : Nat) : String -> Bool =
    let mutable buffer = replicate n 'a'
    let rec loop = function
        | []      -> buffer.next.bead = 'b'
        | x :: xs -> buffer <- buffer.next
                     buffer.bead <- x
                     loop xs
    explode >> loop

(*

> let accept3 = accept 3 ;;
val accept3 : (String -> bool)

> accept3 "bb" ;;
val it : bool = false

> accept3 "bb" ;;
val it : bool = true

> accept3 "aa" ;;  
val it : bool = true

> accept3 "aa" ;;
val it : bool = false

*)   
