(*
Iterative search (linear and binary search) and in-place sorting
algorithms (insertion and selection sort).

fsharpi Mini.fs Effects/Sort.fs
open Effects.Sort ;;
*)

module Effects.Sort

open Mini

// Suchen.

let linearSearch (key : 'elem) (a : Array<'elem>) : Option<Int> =
  let mutable i = 0
  while i < a.Length && a.[i] < key do
    i <- i + 1
  if i < a.Length && a.[i] = key then Some i
                                 else None

let ternarySearch (key : 'elem) (a : Array<'elem>) : Option<Int> =
  let mutable l  =  0
  let mutable u  =  a.Length - 1
  let mutable found = None
  while l <= u && found = None do
    let m = (l + u) / 2
    if    key < a.[m]  then  u <- m - 1
    elif  key = a.[m]  then  found <- Some m
    (* key > a.[m] *)  else  l <- m + 1
  found

// Sortieren.

let insertionSort (a : Array<'elem>) =
  for i in a.Length - 2 .. -1 .. 0 do
    let key = a.[i]
    // insert key into the sorted 
    //   sub-array a.[i + 1], .., a.[a.length - 1]
    let mutable j = i + 1
    while j < a.Length && key > a.[j] do
      a.[j - 1] <- a.[j]
      j <- j + 1
    a.[j - 1] <- key

let selectionSort (a : Array<'elem>) =
  let n = a.Length
  for i in 0 .. n - 2 do
    // determine minimum of a.[i], .., a.[n - 1]
    let mutable m = i
    for j in i + 1 .. n - 1 do
      if a.[j] < a.[m] then
        m <- j;
    // swap a.[i] and a.[m]
    let tmp = a.[i]
    a.[i] <- a.[m]
    a.[m] <- tmp

let countingSort (m : Int, array : Array<Int>) : Array<Int> =
  let counts  =  [|for i in 0 .. m -> 0|]
  for j in array do
    counts.[j] <- counts.[j] + 1
  [|for i in 0 .. m do
      for c in 1 .. counts.[i] do
        yield i|]

// countingSort (6, [|0; 3; 1; 2; 1; 5; 3; 0; 0; 1|]) ;;

