(*
Tracing.

fsharpc Mini.fs Effects/Trace.fs
mono Trace.exe

fsharpi Mini.fs Effects/Trace.fs
open Effects.Trace ;;
*)

module Effects.Trace
#nowarn "62"

open Mini

let Call' (f : 'a -> 'b) : 'a -> 'b =
  fun x -> putline ("call   " ^ show x); f x

let Return' (x : 'a) : 'a =
  putline ("return " ^ show x); x

// turn on and off
let private trace = ref false

let traceOn () =
    trace := true

let traceOff () =
    trace := false

let traceline (s : String) =
    if !trace then putline s // else ()

let Call (f : 'a -> 'b) : 'a -> 'b =
  fun x -> traceline ("call   " ^ show x); f x

let Return (x : 'a) : 'a =
  traceline ("return " ^ show x); x

let rec factorial (n : Nat) : Nat =
  if n = 0N then Return 1N 
            else Return (Call factorial (n - 1N) * n)

// Call factorial 10N ;;

let rec fibonacci (n : Nat) : Nat =
  Return (if n <= 1N then n
                     else Call fibonacci (n - 1N) + Call fibonacci (n - 2N))

// Call fibonacci 5N ;;

let main () =
    traceOn ()
    print (Call factorial 10N)
    traceOff ()
    print (factorial 100N)
    traceOn ()
    print (Call fibonacci 4N)
