(*
UPN calculator (aka reverse Polish notation).

fsharpc Mini.fs Datatypes/Person.fs Effects/Input.fs Effects/UPNCalculator.fs
mono UPNCalculator.exe

fsharpi Mini.fs Datatypes/Person.fs Effects/Input.fs Effects/UPNCalculator.fs 
open Effects.UPNCalculator ;;
upn_calculator () ;;
*)

module Effects.UPNCalculator

open Mini
open Effects

exception Pop
exception Top

// there is no function to empty the stack
module Stack =

  let private stack = ref []

  let push (x : Nat) =
    stack := x :: !stack

  let isempty () : Bool =
    match !stack with
    | []      -> true   
    | x :: xs -> false

  let pop () : Nat =
    match !stack with
    | []      -> raise Pop
    | x :: xs -> stack := xs; x

  let top () : Nat =
    match !stack with
    | []      -> raise Top
    | x :: xs -> x

let upn_calculator () =
    let rec read_eval_print_loop () =
      try
        match Input.query "UPN > " with
        | ""  -> ()
        | "+" -> Stack.push (Stack.pop () + Stack.pop ())
        | "*" -> Stack.push (Stack.pop () * Stack.pop ())
        | "." -> print (Stack.top ())
        | "exit" | "halt" | "q" | "quit" | "stop"
              -> raise EOF
        | s   -> Stack.push (readNat s)
      with
      |  Pop | Top -> putline "stack is empty"
      |  Read      -> putline "enter a number or an operator"
      read_eval_print_loop ()
    in
      try
        read_eval_print_loop ()
      with
      | EOF -> putline "bye bye"

let main = upn_calculator ()
