module Grammar.AcceptorDyck

open Mini
open Grammar.Alphabet

(*

fsharpc Mini.fs Grammar/Alphabet.fs Grammar/AcceptorDyck.fs

Akzeptoren für die Klammernsprache

  { a^n b^n | n <- N }

bzw.

  E = rec x -> eps | a x b

*)

type Follow = List<Alphabet> -> Bool

// Ad-hoc Akzeptor.

let accept_dyck1 : Follow =
  let rec accept_bs (k : int) : Follow = fun input ->
      match input with
      |  []        -> k = 0
      |  A :: rest -> false
      |  B :: rest -> k > 0 && accept_bs (k - 1) rest

  and accept_E_bs (k : int) : Follow = fun input ->
      match input with
      | []        -> k = 0
      | A :: rest -> accept_E_bs (k + 1) rest
      | B :: rest -> k > 0 && accept_bs (k - 1) rest
  in
    accept_E_bs 0

// Mit einem Folgeakzeptor parametrisierter Akkzeptor.

let accept_b (follow : Follow) : Follow = fun input ->
    match input with
    | B :: rest -> follow rest
    | _         -> false

let rec accept_E' (follow : Follow) : Follow = fun input ->
    match input with
    | A :: rest -> accept_E' (accept_b follow) rest
    | _         -> follow input

let end_of_input = fun input ->
    match input with
    | []     -> true
    | _ :: _ -> false

let accept_dyck2 : Follow =
  accept_E' end_of_input

// Systematische Konstruktion.

let symbol (a : Alphabet) (follow : Follow) : Follow = fun input ->
    match input with
    | []        -> false
    | b :: rest -> a = b && follow rest

let rec accept_E (follow : Follow) : Follow =
    fun input ->
         follow input
      || symbol A (accept_E (symbol B follow)) input

let accept_dyck3 : Follow =
  accept_E end_of_input

// Test.

let test (s : String) : Bool =
  accept_dyck3 (Alphabet.lex (explode s))
