(*

fsharpi Mini.fs Grammar/Alphabet.fs Grammar/Regular.fs Grammar/Interpreter.fs

fsharpc Mini.fs Grammar/Alphabet.fs Grammar/Regular.fs Grammar/Interpreter.fs
mono Interpreter.exe

*)

module Grammar.Interpreter

open Mini
open Grammar.Alphabet
open Grammar.Regular

// examples

let any = Alt (Sym A, Sym B)
let r00 = Cat (Rep any, Cat (Sym B, any))
let second_but_last_is_a_B
  = Cat (Rep Any, Cat (Sym B, Any))
let even_no_of_As
  = Rep (Alt (Cat (Sym A, Cat (Rep (Sym B), Sym A)), Sym B))

// a regular expression matcher using derivatives

// is Eps contained in the language

//let rec nullable (reg : Reg<'a>) : Bool =
//    match reg with
let rec nullable = function
    | Eps          -> true
    | Sym _        -> false
    | Cat (r1, r2) -> nullable r1 && nullable r2
    | Empty        -> false
    | Alt (r1, r2) -> nullable r1 || nullable r2
    | Rep r        -> true
    | Any          -> false

// factors (aka derivatives)

// let rec divide (reg : Reg<'a>, x : 'a) : Reg<'a> when 'a : equality =
let rec divide (reg : Reg<'a>, x : 'a) : Reg<'a> =
    match reg with
    | Eps          -> Empty
    | Sym a        -> if x = a then Eps else Empty
    | Cat (r1, r2) -> if nullable r1 then alt (cat (divide (r1, x), r2), divide (r2, x))
                                     else cat (divide (r1, x), r2)
    | Empty        -> Empty
    | Alt (r1, r2) -> alt (divide (r1, x), divide (r2, x))
    | Rep r        -> cat (divide (r, x), Rep r)
    | Any          -> Eps

// divide (second_but_last_is_a_B, A)
// divide (second_but_last_is_a_B, B)

let rec genericAccept (reg : Reg<'a>, input : List<'a>) : Bool =
    match input with
    | []        -> nullable reg
    | a :: rest -> genericAccept (divide (reg, a), rest)

// genericAccept (second_but_last_is_a_B, A :: A :: A :: [])
// genericAccept (second_but_last_is_a_B, A :: B :: A :: [])
// genericAccept (even_no_of_As, A :: B :: A :: [])
// genericAccept (even_no_of_As, A :: B :: B :: [])

//let accept_r00 (input : List<Alphabet>) : Bool =
//  genericAccept (r00, input)
let accept_r00 = fun input -> genericAccept (r00, input)

let test (s : String) : Bool =
  accept_r00 (lex (explode s))

// test "aaa"
// test "aba"

let main =
    print (test "aaa")
    print (test "aba")
