(*
   
fsharpc Grammar/Regular.fs

*)

module Grammar.Regular

// - uses parametric types and polymorphic functions
// - additional constructor: Any

// Reguläre Ausdrücke.

type Reg<'a> =
    | Eps
    | Sym of 'a
    | Cat of Reg<'a> * Reg<'a>
    | Empty
    | Alt of Reg<'a> * Reg<'a>
    | Rep of Reg<'a>
    | Any

// Clevere Konstruktoren, die einfache algebraische Eigenschaften
// ausnutzen.

let cat (reg1 : Reg<'a>, reg2 : Reg<'a>) : Reg<'a> =
    match (reg1, reg2) with
    | (Eps,   r) | (r, Eps)   -> r
    | (Empty, r) | (r, Empty) -> Empty
    | _                       -> Cat (reg1, reg2)

let alt (reg1 : Reg<'a>, reg2 : Reg<'a>) : Reg<'a> =
    match (reg1, reg2) with
    | (Empty, r) | (r, Empty) -> r
    | _                       -> Alt (reg1, reg2)
