(*

Natürliche Zahlen und arithmetische Ausdrücke.

fsharpi Mini.fs Interpreter/Naturals.fs

open Interpreter.Naturals ;;

*)

module Interpreter.Naturals

// abstrakte Syntax

type Expr =
    | Num of Nat
    | Add of Expr * Expr

// metazirkulärer Interpreter

type Value =
    Nat

let rec evaluate : Expr -> Value = function
    | Num n        -> n
    | Add (e1, e2) -> evaluate e1 + evaluate e2

// abstrakte Maschine

type Frame =
    | Add1 of Expr
    | Add2 of Value

type Stack =
    Frame list

type State =
    | Ret  of Value * Stack
    | Eval of Expr  * Stack

let step : State -> State = function
// Auswertung
    | Eval (Num n,        stack) -> Ret (n, stack)
    | Eval (Add (e1, e2), stack) -> Eval (e1, Add1 e2 :: stack)
// Rückgabe
    | Ret (v, [])                -> Ret  (v, [])
    | Ret (v1, Add1 e2 :: stack) -> Eval (e2, Add2 v1 :: stack)
    | Ret (v2, Add2 v1 :: stack) -> Ret  (v1 + v2, stack)

let eval (e : Expr) : Value =
    let rec loop = function
        | Ret  (v, []) -> v
        | state        -> loop (step state)
    loop (Eval (e, []))

// Beispiele

let ex1 = Add (Num 4N, Num 7N)
let ex2 = Add (Num 1N, Num 2N)
let ex3 = Add (ex1, ex2)

// Eval (ex3, []) ;;
// step it ;;

let rec leftist n =
    if n = 0N then Num 0N
              else Add (leftist (n - 1N), Num n)

let rec rightist n =
    if n = 0N then Num 0N
              else Add (Num n, rightist (n - 1N))

// Eval (leftist  5N, []) ;;
// Eval (rightist 5N, []) ;;
// step it ;;
// eval (leftist  1000N) ;;
// eval (rightist 1000N) ;;

///////////////////////////////////////////////////////////////////////////////

// experimentell

let rec compile : Expr * Stack -> Value * Stack = function
    | (Num n,        stack) -> (n, stack)
    | (Add (e1, e2), stack) -> compile (e1, Add1 e2 :: stack)

let execute : Value * Stack -> Value * Stack = function
    | (v, [])                -> (v, [])
    | (v1, Add1 e2 :: stack) -> compile (e2, Add2 v1 :: stack)
    | (v2, Add2 v1 :: stack) -> (v1 + v2, stack)
