(*
Objects versus recursive records.

fsharpi Mini.fs Objects/Clock.fs
open Objects.Clock ;;
*)

module Objects.Clock

open Mini

module Record =
    type Clock =
        {
            Tick : Nat -> Nat
            Time : Unit -> Nat
        }

    let clock () =
        let mutable hours = 0N
        let rec self =
            {
                Tick = fun inc ->
                         hours <- hours + inc
                         self.Time ()
                Time = fun () ->
                         hours % 24N
            }
        self

(*
> open Objects.Clock.Record ;;
> let c = clock () ;;
> c.Tick (1N) ;;
val it : Nat = 1N
> c.Tick (2N) ;;
val it : Nat = 3N
> c.Tick (3N) ;;
val it : Nat = 6N
> c.Time () ;;
val it : Nat = 6N
*)   

module MutableRecord =
    type Clock =
        {
            mutable hours : Nat
            Tick : Nat -> Nat
            Time : Unit -> Nat
        }

    let clock () =
        let rec self =
            {
                hours = 0N
                Tick = fun inc ->
                         self.hours <- self.hours + inc
                         self.Time ()
                Time = fun () ->
                         self.hours % 24N
            }
        self

module Object =
    type Clock =
        interface
            abstract Tick : Nat -> Nat
            abstract Time : Nat with get
         end

    let clock () =
        let mutable hours = 0N
        { new Clock with
            member self.Tick inc =
                     hours <- hours + inc
                     self.Time
            member self.Time
                     with get () = hours % 24N
        }

(*
> open Objects.Clock.Object ;;
> let c = clock () ;;
> c.Tick 1N ;;
val it : Nat = 1N
> c.Tick 2N ;;
val it : Nat = 3N
> c.Tick 3N ;;
val it : Nat = 6N
> c.Time ;;
val it : Nat = 6N
*)   
