(*
Mutable finite maps aka dictionaries.

fsharpi Mini.fs Algorithmics/Tree.fs Algorithmics/Map/SearchTree.fs Objects/Dictionary.fs
open Objects.Dictionary ;;
*)

module Objects.Dictionary
#nowarn "62"

open Mini
open Algorithmics.Tree
open Algorithmics.Map
open System.Collections
open System.Collections.Generic

let enumerateKeys (Rep tree) =
    let rec inorder = function
        | Leaf           -> Seq.empty
        | Node (l, x, r) -> seq { yield! inorder l; yield x.key; yield! inorder r}
    inorder tree

type Dictionary<'key, 'value when 'key : comparison> =
    interface
        abstract member Add  : 'key * 'value -> unit
        abstract member Item : 'key -> 'value with get, set
    end

type SearchTree<'key, 'value when 'key : comparison> () =
    class
        let mutable name = "<unnamed>"
        let mutable mapping : Map<'key, 'value> = empty

        member self.Name
            with get () = name
            and  set s  = name <- s
        
        member self.Add (key, value) =
            mapping <- add (key, value) mapping

        member self.Lookup key =
            match lookup key mapping with
            | None       -> raise (KeyNotFoundException ())
            | Some value -> value

        member self.Item
            with get key       = self.Lookup key
            and  set key value = self.Add (key, value)

        interface Dictionary<'key, 'value> with
            member self.Add (key, value) =
                self.Add (key, value)
            member self.Item
                with get key       = self.[key]
                and  set key value = self.[key] <- value

        interface IEnumerable<'key> with
            member self.GetEnumerator () = (enumerateKeys mapping).GetEnumerator ()
        interface IEnumerable with 
            member self.GetEnumerator () = (self :> IEnumerable<'key>).GetEnumerator () :> IEnumerator
    end

(*
let perfume = SearchTree<Nat, String> () ;;
perfume.Add (4711N, "Kölnisch Wasser") ;;
perfume.Add (5N, "Chanel") ;;
perfume.Lookup 5N ;;
perfume.[5N] ;;
perfume.[0N] <- "Eau" ;;
perfume.[0N] ;;
perfume.[0N] <- "Wasser" ;;
perfume.[0N] ;;
for key in perfume do putline perfume.[key] ;;
*)

(*
let phoneBook = SearchTree<String, Nat> () ;;
phoneBook.Name <- "Telefonbuch" ;;
phoneBook.Name ;;
phoneBook.Add ("Harry", 4711N) ;;
phoneBook.Add ("Lisa", 815N) ;;
phoneBook.["Harry"] ;;
phoneBook.["Harry"] <- 47110N;;
phoneBook.["Harry"] ;;
for key in phoneBook do putline (key ^ " |-> " ^ show phoneBook.[key]) ;;
for key in phoneBook do putline (key ^ "'s phone number is " ^ show phoneBook.[key]) ;;
*)
