(*
Sequence expressions: generating all permutations.

fsharpi Mini.fs Objects/Permutations.fs
open Objects.Permutations ;;
*)

module Objects.Permutations

open Mini

// using lists of lists

let rec deletions = function
  | []      -> []
//  | x :: xs -> (x, xs) :: [ for (y, ys) in deletions xs do yield (y, x :: ys) ]
  | x :: xs -> [ yield (x, xs); for (y, ys) in deletions xs do yield (y, x :: ys) ]

let rec dpermutations = function
  | [] -> [[]]
  | xs -> [ for (y, ys) in deletions xs do for zs in dpermutations ys do yield y :: zs ]

let rec insertions x = function
  | []             -> [ [x] ]
  | xs & (y :: ys) -> [ yield (x :: xs); for zs in insertions x ys do yield y :: zs ]

let rec ipermutations = function
  | []      ->  [ [] ]
  | x :: xs ->  [ for ys in ipermutations xs do for zs in insertions x ys do yield zs ]

// using sequences of lists

let rec deletions' = function
  | []      -> Seq.empty
  | x :: xs -> seq { yield (x, xs); for (y, ys) in deletions' xs do yield (y, x :: ys) }

let rec dpermutations' = function
  | [] -> seq { yield [] }
  | xs -> seq { for (y, ys) in deletions' xs do for zs in dpermutations' ys do yield y :: zs }

let rec insertions' x = function
  | []             -> seq { yield [x] }
  | xs & (y :: ys) -> seq { yield (x :: xs); for zs in insertions' x ys do yield y :: zs }

let rec ipermutations' = function
  | []      ->  seq { yield [] }
  | x :: xs ->  seq { for ys in ipermutations' xs do for zs in insertions' x ys do yield zs }

// 3.628.800 elements:
// dpermutations [1 .. 10] ;;
// dpermutations' [1 .. 10] ;;

let lightningSort = List.sort // >> List.rev

let test n =
    for i in 0 .. n do
        for xs in dpermutations' [1 .. i] do
            if lightningSort xs <> List.sort xs then
                raise (Panic "Harry!")

let inputs n =
    [ for i in 0 .. n do
        yield! dpermutations' [1 .. i] ]
