(*
Inheritance breaks abstraction (subclasses versus delegation).

fsharpi Mini.fs Objects/StepCounter.fs
open Objects.StepCounter ;;
*)

module Objects.StepCounter

open Mini

module Classes =
    
  type StepCounter () =
      class
          let mutable n = 0

          abstract member Inc : Unit -> Unit
          default self.Inc () =
              putline "Inc:  base class"
              n <- n + 1

          abstract member Step : Int -> Unit
          default self.Step k =
              putline "Step: base class"
              n <- n + k
// OR              
//              for i in 1 .. k do
//                  self.Inc ()
      end

  type ParityCounter () =
      class
          inherit StepCounter ()

          let mutable even = true

          override self.Inc () =
              putline "Inc:  subclass"
              base.Inc ()
              even <- not even
(*
          override self.Step k =
              putline "Step: subclass"
              base.Step k
              even <- even = (k % 2 = 0)
*)
          member self.Parity = even
      end

// let counter = Classes.ParityCounter () ;;
// counter.Step 3 ;;
// counter.Parity ;;

module Objects =
    
  type IStepCounter =
    interface
        abstract member Inc  : Unit -> Unit
        abstract member Step : Int  -> Unit
    end

  let stepCounter () =
      let mutable n = 0
      {
          new IStepCounter with

          member self.Inc () =
              putline "Inc:  delegatee"
              n <- n + 1

          member self.Step k =
              putline "Step: delegatee"
              n <- n + k
// OR          
//              for i in 1 .. k do
//                  self.Inc ()
      }

  type IParityCounter =
    interface
        inherit IStepCounter
        abstract member Parity : Bool
    end

  let parityCounter () =
      let basic = stepCounter ()
      let mutable even = true
      {
        new IParityCounter with
          member self.Inc () =
              putline "Inc:  delegator"
              basic.Inc ()
              even <- not even

          member self.Step k =
              putline "Step: delegator"
              basic.Step k
              even <- even = (k % 2 = 0)

          member self.Parity = even
      }

// let counter = Objects.parityCounter () ;;
// counter.Step 3 ;;
// counter.Parity ;;
