(*
Template design pattern (classical variant using abstract classes and
subclasses, and another one using HOF).

fsharpc Mini.fs Datatypes/Person.fs Effects/Input.fs Objects/Template.fs
mono Template.exe

DOES NOT WORK: * Assertion at mini-trampolines.c:705 if a non-number is entered
fsharpi Mini.fs Datatypes/Person.fs Effects/Input.fs Objects/Template.fs
open Mini ;;
open Objects.Template ;;
*)

module Objects.Template
#nowarn "62"

open Mini
open Effects.Input

[<AbstractClass>]
type TwoPlayerGame () =
    class
        abstract member Finished : Bool
        abstract member Position : Unit -> Unit
        abstract member Move     : Bool -> Unit
        abstract member Winner   : Bool -> Unit

        member self.Play () =
            try
                let mutable turn = false
                while not self.Finished do
                    self.Position ()
                    turn <- not turn
                    self.Move turn
                self.Winner turn
            with
            | EOF -> putline "\nbye bye"
    end

let bars (i : Nat)   = String.replicate (int i) "|"
let (^^) x (i : Nat) = List.replicate (int i) x

type Nim (n : Nat) =
//    class
        inherit TwoPlayerGame ()

        let mutable matches = n

        let announce i =
            putline ("I take " ^ i.ToString ()); i

        override self.Finished =
            matches = 0N

        override self.Position () =
//            putline (show matches ^ " matches remain")
            putline (String.concat " " (((bars 5N) ^^ (matches / 5N)) @ [bars (matches % 5N)]))
(*            
            for i in 1N .. matches do
                putchar '|'
                if i % 5N = 0N && i < n then 
                    putchar ' '
            putchar '\n'
*)

        override self.Move (human : Bool) : Unit =
            let i =
                if human then
                    checkedQuery ("number of matches", both (isNat, both (isGreater 0N, isLess 4N)))
                else
                    announce (max 1N (matches % 4N))
            matches <- matches - i

        override self.Winner human =
            putline ((if human then "you" else "I") ^ " win")
//    end

// open Objects.Template ;;
// (Nim 10N).Play () ;;

let main = (Nim 10N).Play ()

module HOF =

  type ITwoPlayerGame =
      interface
          abstract member Finished : Bool
          abstract member Position : Unit -> Unit
          abstract member Move     : Bool -> Unit
          abstract member Winner   : Bool -> Unit
      end

  let play (game : ITwoPlayerGame) =
      try
          let mutable turn = false
          while not game.Finished do
              game.Position ()
              turn <- not turn
              game.Move turn
          game.Winner turn
      with
      | EOF -> putline "\nbye bye"

  let nim (n : Nat) =
      let mutable matches = n
      let announce i = putline ("I take " ^ i.ToString ()); i
      { new ITwoPlayerGame with
        member self.Finished =
            matches = 0N

        member self.Position () =
            putline (String.concat " " (((bars 5N) ^^ (matches / 5N)) @ [bars (matches % 5N)]))

        member self.Move (human : Bool) : Unit =
            let i =
                if human then
                    checkedQuery ("number of matches", both (isNat, both (isGreater 0N, isLess 4N)))
                else
                    announce (max 1N (matches % 4N))
            matches <- matches - i

        member self.Winner human =
            putline ((if human then "you" else "I") ^ " win") }

let main2 () = HOF.play (HOF.nim 10N)
