(*
Generic interfaces: stacks, UPN calculator.

fsharpi Mini.fs Datatypes/Person.fs Effects/Input.fs Objects/UPNCalculator.fs
open Objects.UPNCalculator ;;
*)

module Objects.UPNCalculator

open Mini
open Effects

type IStack<'elem> =
  interface
    abstract member IsEmpty : Bool
    abstract member Push    : 'elem -> Unit
    abstract member Pop     : 'elem
    abstract member Top     : 'elem
  end

exception Pop
exception Top

let stack<'elem> () : IStack<'elem> =
    let mutable stack = []
    {
      new IStack<'elem> with
        member self.IsEmpty =
          List.isEmpty stack

        member self.Push (x : 'elem) =
          stack <- x :: stack

        member self.Pop : 'elem =
          match stack with
          | []      -> raise Pop
          | x :: xs -> stack <- xs ; x

        member self.Top : 'elem =
          match stack with
          | []      -> raise Top
          | x :: xs -> x
    }

let upn_calculator_deluxe () =
    let stacks : IStack<IStack<Nat>> = stack ()

    try
      stacks.Push (stack ())
      while true do
        try
          match Input.query "UPN > " with
          | ""  -> ()
          | "+" -> stacks.Top.Push (stacks.Top.Pop + stacks.Top.Pop)
          | "*" -> stacks.Top.Push (stacks.Top.Pop * stacks.Top.Pop)
          | "." -> putline (stacks.Top.Top.ToString ())
          | "exit" | "halt" | "q" | "quit" | "stop"
                -> raise EOF
          | "(" -> stacks.Push (stack ())
          | ")" -> stacks.Pop |> ignore
          | s   -> stacks.Top.Push (readNat s)
        with
        |  Pop | Top -> putline "stack is empty"
        |  Read      -> putline "enter a number or an operator"
      with
      | EOF -> putline "bye bye"

let main = upn_calculator_deluxe ()

// generic class (otherwise same as above)

type Stack<'elem> () =
  class
    let mutable stack = []

    member self.IsEmpty =
      List.isEmpty stack

    member self.Push (x : 'elem) =
      stack <- x :: stack

    member self.Pop : 'elem =
      match stack with
      | []      -> raise Pop
      | x :: xs -> stack <- xs ; x

    member self.Top : 'elem =
      match stack with
      | []      -> raise Top
      | x :: xs -> x
  end
